/*
	crc.c
	Sunday, March 5, 1995 6:21:30 PM

	CRC Checksum generation for a file.

Aug 15, 2000 (Loren Petrich):
	Using object-oriented file handler

Nov 6, 2000 (Loren Petrich);
	Suppressed commented-out FileSpecifier function
*/

#include <stdlib.h>

#include "cseries.h"
#include "FileHandler.h"
#include "crc.h"

#ifdef env68k
	#pragma segment file_io
#endif

/* ---------- constants */
#define TABLE_SIZE (256)
#define CRC32_POLYNOMIAL 0xEDB88320L
#define BUFFER_SIZE 1024

/* ---------- local data */
static uint32 *crc_table= NULL;

/* ---------- local prototypes ------- */
static uint32 calculate_file_crc(unsigned char *buffer, 
	short buffer_size, OpenedFile& OFile);
static uint32 calculate_buffer_crc(long count, uint32 crc, void *buffer);
static bool build_crc_table(void);
static void free_crc_table(void);

/* -------------- Entry Point ----------- */
uint32 calculate_crc_for_file(FileSpecifier& File)
{
	uint32 crc = 0;
	
	OpenedFile OFile;
	if (File.Open(OFile))
	{
		crc= calculate_crc_for_opened_file(OFile);
		OFile.Close();
	}
	
	return crc;
}

uint32 calculate_crc_for_opened_file(OpenedFile& OFile)
{
	uint32 crc = 0;
	unsigned char *buffer;

	/* Build the crc table */
	if(build_crc_table())
	{
		buffer = new byte[BUFFER_SIZE];
		if(buffer) 
		{
			crc= calculate_file_crc(buffer, BUFFER_SIZE, OFile);
			delete []buffer;
		}
		
		/* free the crc table! */
		free_crc_table();
	}

	return crc;
}

/* Calculate the crc for a file using the given buffer.. */
uint32 calculate_data_crc(
	unsigned char *buffer,
	long length)
{
	uint32 crc = 0;

	assert(buffer);
	
	/* Build the crc table */
	if(build_crc_table())
	{
		/* The odd permutions ensure that we get the same crc as for a file */
		crc = 0xFFFFFFFFL;
		crc = calculate_buffer_crc(length, crc, buffer);
		crc ^= 0xFFFFFFFFL;

		/* free the crc table! */
		free_crc_table();
	}

	return crc;
}

/* ---------------- Private Code --------------- */
static bool build_crc_table(
	void)
{
	bool success= false;

	assert(!crc_table);
	crc_table= new uint32[TABLE_SIZE];
	if(crc_table)
	{
		/* Build the table */
		short index, j;
		uint32 crc;

		for(index= 0; index<TABLE_SIZE; ++index)
		{
			crc= index;
			for(j=0; j<8; j++)
			{
				if(crc & 1) crc=(crc>>1) ^ CRC32_POLYNOMIAL;
				else crc>>=1;
			}
			crc_table[index] = crc;
		}
		
		success= true;
	}
	
	return success;
}

static void free_crc_table(
	void)
{
	assert(crc_table);
	delete []crc_table;
	crc_table= NULL;
}

/* Calculate for a block of data incrementally */
static uint32 calculate_buffer_crc(
	long count, 
	uint32 crc, 
	void *buffer)
{
	unsigned char *p;
	uint32 a;
	uint32 b;

	p= (unsigned char *) buffer;
	while (count--) 
	{
		a= (crc >> 8) & 0x00FFFFFFL;
		b= crc_table[((int) crc ^ *p++) & 0xff];
		crc= a^b;
	}
	return crc;
}

/* Calculate the crc for a file using the given buffer.. */
static uint32 calculate_file_crc(
	unsigned char *buffer, 
	short buffer_size,
	OpenedFile& OFile)
{
	uint32 crc;
	long count;
	// FileError err;
	long file_length, initial_position;
	
	/* Save and restore the initial file position */
	assert(OFile.GetPosition(initial_position));
	// initial_position= get_fpos(refnum);

	/* Get the file_length */
	assert(OFile.GetLength(file_length));
	// file_length= get_file_length(refnum);
	
	/* Set to the start of the file */
	assert(OFile.SetPosition(0));
	// set_fpos(refnum, 0l);

	crc = 0xFFFFFFFFL;
	while(file_length) 
	{
		if(file_length>buffer_size)
		{
			count= buffer_size;
		} else {
			count= file_length;
		}

		assert(OFile.Read(count,buffer));
		// err= read_file(refnum, count, buffer);
		// vassert(!err, csprintf(temporary, "Error: %d", err));
		
		crc = calculate_buffer_crc(count, crc, buffer);
		file_length -= count;
	}
	
	/* Restore the file position */
	assert(OFile.SetPosition(initial_position));
	// set_fpos(refnum, initial_position);

	return (crc ^= 0xFFFFFFFFL);
}
